// Kye Wei
// Life Simulation
// ICS4U

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

import javax.imageio.*;
import java.io.*;

public class LifeSimulation extends JPanel implements ActionListener, MouseMotionListener, MouseListener
{
    public Colony colony;
    public JButton button0;
    public JButton button1;
    public JButton button2;
    public JButton button3;
    public JButton button4;
    public JButton button5;
    public JButton button6;

    public double dens = 0.6; // density used when making new palette
    public double replacedens = 0.6; // density used in "repopulate"
    int eradicatex = 0; // 0= add dots, -1= remove dots, 1 - eradicate, 2-repopulate
    boolean makenew = true;
    double int1 = 0, int2 = 0, int3 = 0; // starting circle radius position, namely x, y, diameter
    public int x1; // more circle positions
    public int y1;

    public Timer t; // timer
    public LifeSimulation ()
    {
	super ();

	addMouseMotionListener (this);
	addMouseListener (this);

	setLayout (new BorderLayout ());

	JRadioButton rd1 = new JRadioButton ("Slow"); // add slow button
	rd1.setActionCommand ("Slow");
	rd1.addActionListener (this);
	rd1.setToolTipText ("1 Generation Per 200 milliseconds");
	JRadioButton rd2 = new JRadioButton ("Medium"); // add medium button
	rd2.setActionCommand ("Medium");
	rd2.setSelected (true);
	rd2.addActionListener (this);
	rd2.setToolTipText ("1 Generation Per 50 milliseconds");
	JRadioButton rd3 = new JRadioButton ("Fast"); // add fast button
	rd3.setActionCommand ("Fast");
	rd3.addActionListener (this);
	rd3.setToolTipText ("1 Generation Per 10 milliseconds");

	ButtonGroup group1 = new ButtonGroup (); // group the three radio buttons
	group1.add (rd1);
	group1.add (rd2);
	group1.add (rd3);

	JRadioButton rd9 = new JRadioButton ("Add/Remove Individual Life"); // button to switch to add dot mode
	rd9.setActionCommand ("Individual");
	rd9.setSelected (true);
	rd9.addActionListener (this);
	rd9.setToolTipText ("Click adds life to selected square");

	JRadioButton rd11 = new JRadioButton ("Add/Remove Area Life"); // button to switch to add circle mode
	rd11.setActionCommand ("Area");
	rd11.addActionListener (this);
	rd11.setToolTipText ("Left click and drag repopulates remaining empty spaces life with selected density, right click and drag eradicates life with a 20% survival rate");

	ButtonGroup group2 = new ButtonGroup (); // group buttons

	group2.add (rd9);
	group2.add (rd11);

	button0 = new JButton ("Advance"); // advance button
	button0.addActionListener (this);
	button0.setActionCommand ("Advance");
	button0.setToolTipText ("Advance by one generation");

	button1 = new JButton ("Run Simulation"); // run button
	button1.addActionListener (this);
	button1.setActionCommand ("Run");
	button1.setToolTipText ("Run automated simulation with selected speed");

	button2 = new JButton ("Stop Simulation"); //stop button
	button2.addActionListener (this);
	button2.setActionCommand ("Stop");
	button2.setEnabled (false);
	button2.setToolTipText ("Stop automated simulation");

	button3 = new JButton ("Reset"); // reset palette
	button3.addActionListener (this);
	button3.setActionCommand ("Reset");
	button3.setToolTipText ("Generates new palette with selected density (adjustments on the left)");

	button4 = new JButton ("Gosper Glider Gun O:"); // adds a predefined gosper glider gun palette
	button4.addActionListener (this);
	button4.setActionCommand ("pew");
	button4.setToolTipText ("Generates new palette with the built-in Gosper Glider Gun");

	button5 = new JButton ("Save"); // saves to file
	button5.addActionListener (this);
	button5.setActionCommand ("Save");
	button5.setToolTipText ("Saves current palette to file; file contains (first line)name of design, (following lines)coordinates (row, column) of points where life exists");

	button6 = new JButton ("Load"); // loads using one of two formats
	button6.addActionListener (this);
	button6.setActionCommand ("Load");
	button6.setToolTipText ("Loads predefined palettes from file; accepts either files with . and * with line breaks for each row, or files with coordinates of points where life exists");

	JSlider slider1 = new JSlider (JSlider.VERTICAL, 0, 100, 60);
	slider1.setBorder (BorderFactory.createTitledBorder ("Density %")); // palette density when making new one
	slider1.setMajorTickSpacing (20);
	slider1.setMinorTickSpacing (5);
	slider1.setPaintTicks (true);
	slider1.setPaintLabels (true);
	slider1.setPreferredSize (new Dimension (80, 400));
	slider1.setToolTipText ("Adjusts the population density used by the reset function");

	slider1.addChangeListener (new ChangeListener ()
	{
	    public void stateChanged (ChangeEvent e)
	    {
		JSlider source = (JSlider) e.getSource ();

		if (!source.getValueIsAdjusting ())
		{
		    dens = source.getValue () * 0.01; // since value is 1-100, we need 0-1 so we need to *0.01
		    colony = new Colony (dens);
		}
	    }
	}
	);

	JSlider slider2 = new JSlider (JSlider.HORIZONTAL, 0, 100, 60);
	slider2.setBorder (BorderFactory.createTitledBorder ("\"Repopulate\" Density % (Good range is 40-60%)")); //overlayed density for circle
	slider2.setMajorTickSpacing (20);
	slider2.setMinorTickSpacing (5);
	slider2.setPaintTicks (true);
	slider2.setPaintLabels (true);
	slider2.setPreferredSize (new Dimension (400, 60));
	slider2.setToolTipText ("Adjusts the population density used by the repopulate function");
	slider2.addChangeListener (new ChangeListener ()
	{
	    public void stateChanged (ChangeEvent e)
	    {
		JSlider source = (JSlider) e.getSource ();

		if (!source.getValueIsAdjusting ())
		{
		    replacedens = source.getValue () * 0.01; // since value is 1-100, we need 0-1 so we need to *0.01
		}
	    }
	}
	);

	JPanel Buttons1 = new JPanel (); // Make three JFrames to contain the components and add the frames
	Buttons1.add (rd1);
	Buttons1.add (rd2);
	Buttons1.add (rd3);

	Buttons1.add (button0);
	Buttons1.add (button1);
	Buttons1.add (button2);
	Buttons1.add (button3);
	Buttons1.add (button4);
	Buttons1.add (button5);
	Buttons1.add (button6);

	JPanel Buttons2 = new JPanel ();
	Buttons2.add (rd9);
	Buttons2.add (rd11);
	Buttons2.add (slider2);

	add (Buttons1, BorderLayout.NORTH);
	add (slider1, BorderLayout.WEST);
	add (Buttons2, BorderLayout.SOUTH);

	colony = new Colony (dens); // create colony with 60% density
	//Code to add Timer

    }


    public static void main (String[] args)
    {
	JFrame frame = new JFrame ("Life Simulation - 0"); // starting generation #
	frame.setSize (910, 650);
	frame.setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);
	frame.setVisible (true);

	LifeSimulation panel = new LifeSimulation ();

	frame.getContentPane ().add (panel);
	frame.getContentPane ().setSize (900, 640);
	frame.setResizable (true);

	frame.show ();
    }


    public void paintComponent (Graphics g)
    {
	super.paintComponent (g);
	colony.show (g);

	if (!makenew) // eradicate mode requires red.
	{
	    g.setColor (Color.red);
	    g.fillOval ((int) int1, (int) int2, (int) int3, (int) int3);
	}
	else if (makenew) // add mode requires green.
	{
	    g.setColor (Color.green);
	    g.fillOval ((int) int1, (int) int2, (int) int3, (int) int3);
	}
	repaint ();
    }


    public void actionPerformed (ActionEvent event)
    {
	String command = event.getActionCommand ();

	if (command == null || command.equals ("Advance")) // if event is sent by timer, getActionCommand is null.
	{
	    colony.advance (); // update life status
	    repaint ();
	    ((JFrame) getParent ().getParent ().getParent ().getParent ()).setTitle ("Life Simulation - " + colony.generations); // update generations
	}
	else if (command.equals ("Run"))
	{
	    if (t == null)
		t = new Timer (50, this);
	    t.start (); // start simulation

	    button1.setEnabled (false); // cannot run again, can stop, cannot load
	    button2.setEnabled (true);
	    button5.setEnabled (false);
	    button6.setEnabled (false);
	}
	else if (command.equals ("Stop"))
	{
	    t.stop ();
	    button1.setEnabled (true); // cannot stop, can run, can load
	    button2.setEnabled (false);
	    button5.setEnabled (true);
	    button6.setEnabled (true);

	}
	else if (command.equals ("Reset"))
	{
	    try // stop timers if they exist
	    {
		t.stop ();
	    }
	    catch (NullPointerException e)
	    {
	    }

	    button1.setEnabled (true); // cannot stop, can run, can load
	    button2.setEnabled (false);
	    button5.setEnabled (true);
	    button6.setEnabled (true);


	    colony = new Colony (dens);
	}

	else if (command.equals ("pew"))
	{
	    try // stop timers if they exist
	    {
		t.stop ();
	    }
	    catch (NullPointerException e)
	    {
	    }

	    button1.setEnabled (true); // cannot stop, can run, can load
	    button2.setEnabled (false);
	    button5.setEnabled (true);
	    button6.setEnabled (true);
	    colony = new Colony (0.0, "pew");
	}


	else if (command.equals ("Save")) // export
	{
	    try
	    {
		String fileName = JOptionPane.showInputDialog ("Enter file name (include .txt):"); // gets name of file
		String designName = JOptionPane.showInputDialog ("Enter design name (leave blank for \"Untitled\"):"); // gets label of file
		colony.exporter (fileName, designName); // send name and label
	    }
	    catch (NullPointerException e)  // error display message
	    {
		JOptionPane.showMessageDialog ((JFrame) getParent ().getParent ().getParent ().getParent (), "File was not saved.");
	    }
	}


	else if (command.equals ("Load"))
	{ //display message, asks for either mode
	    int reply = JOptionPane.showConfirmDialog ((JFrame) getParent ().getParent ().getParent ().getParent (), "Use <format with . and *> rather than <list of row,column> ?", "Import Method?", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null);
	    try
	    {
		String fileName = JOptionPane.showInputDialog ("Enter file name (include .txt):"); // enter file name
		colony = new Colony (0.0);
		String temp = colony.importer (fileName, reply); // send mode in, make new colony
		if (!temp.equals (""))
		    JOptionPane.showMessageDialog ((JFrame) getParent ().getParent ().getParent ().getParent (), temp);
	    }
	    catch (NullPointerException e)  // error message
	    {
		JOptionPane.showMessageDialog ((JFrame) getParent ().getParent ().getParent ().getParent (), "File was not loaded.");
	    }
	}
	else if (command.equals ("Individual")) // handles new button
	{
	    eradicatex = 0;

	}
	else if (command.equals ("Area")) //handles eradicate button
	{
	    eradicatex = 1;

	}
	else if (command.equals ("Slow") || command.equals ("Medium") || command.equals ("Fast")) // speed adjust
	{
	    boolean start;
	    try //stop current timer
	    {
		start = t.isRunning ();
		t.stop ();
	    }
	    catch (NullPointerException e)
	    {
		start = false;
	    }

	    if (command.equals ("Slow")) //make new timers according to speed
		t = new Timer (200, this);
	    if (command.equals ("Medium"))
		t = new Timer (50, this);
	    if (command.equals ("Fast"))
		t = new Timer (10, this);

	    if (start)
		t.start ();
	}
    }


    public void mousePressed (MouseEvent e)
    {
	int x = e.getX ();
	int y = e.getY ();
	x1 = x;
	y1 = y;


	if (e.getButton () == MouseEvent.BUTTON3)
	{
	    makenew = false;
	}
	else if (e.getButton () == MouseEvent.BUTTON1)
	{
	    makenew = true;
	}


	if (eradicatex == 0 || eradicatex == -1) // either add or remove dot mode
	{
	    int column = (int) Math.round ((x - 90 - 2) / 4); // convert to column
	    int ro = (int) Math.round ((y - 40 - 2) / 4); //convert to row

	    if (ro >= 0 && ro < colony.grid.length && column >= 0 && column < colony.grid [0].length) //if in range
	    {
		if (makenew)
		    colony.grid [ro] [column] = true; // add life if in add mode
		else if (!makenew)
		    colony.grid [ro] [column] = false; //remove live if in remove mode
	    }
	}
    }


    public void mouseReleased (MouseEvent e)
    {
	x1 = 0;
	y1 = 0;

	double xspot = int1 + int3 / 2; //end coordinate spot
	double yspot = int2 + int3 / 2; //end coordinate spot

	//  g.fillRect (col * 4 + 90, row * 4 + 40, 4, 4);   // draw life form

	double column = (xspot - 90 - 2) / 4; //coordinate spot
	double ro = (yspot - 40 - 2) / 4; //coordinate spot

	for (int row = 0 ; row < colony.grid.length ; row++)
	{
	    for (int col = 0 ; col < colony.grid [0].length ; col++)
	    {
		if ((int) Math.sqrt (Math.pow ((row - ro), 2) + Math.pow ((col - column), 2)) <= (int) (int3 / 2 / 4))
		{
		    if (e.getButton () == MouseEvent.BUTTON3)
		    {
			makenew = false;
			if (colony.grid [row] [col] == true) // eradicates life that exists at a 20% success rate
			    colony.grid [row] [col] = (Math.random () < 0.20);
		    }
		    else if (e.getButton () == MouseEvent.BUTTON1)
		    {
			makenew = true;
			if (colony.grid [row] [col] == false) // replaces life that is dead with life at a specified percentage
			    colony.grid [row] [col] = (Math.random () < replacedens); // around 0.5 - 0.6 is best
		    }
		}
	    }
	}
	int1 = 0;
	int2 = 0;
	int3 = 0;
    }


    public void mouseEntered (MouseEvent e)
    {
    }


    public void mouseExited (MouseEvent e)
    {
    }


    public void mouseClicked (MouseEvent e)
    {
    }


    public void mouseMoved (MouseEvent e)
    {
    }


    public void mouseDragged (MouseEvent e)
    {
	int x = e.getX ();
	int y = e.getY ();

	if (e.getButton () == MouseEvent.BUTTON3)
	{
	    makenew = false;
	}
	else if (e.getButton () == MouseEvent.BUTTON1)
	{
	    makenew = true;
	}
	if (eradicatex == 1)
	{
	    double radius = Math.sqrt (Math.pow ((x - x1), 2) + Math.pow ((y - y1), 2)); // continually updates the radius of the circle
	    if (e.getButton () != MouseEvent.BUTTON2)
	    {
		int1 = x1 - radius;  // top left x
		int2 = y1 - radius;  // top left y
		int3 = radius * 2;  //diameter
	    }
	}
	else if (eradicatex == 0) // single block add
	{
	    int column = (int) Math.round ((x - 90 - 2) / 4);
	    int ro = (int) Math.round ((y - 40 - 2) / 4);

	    if (ro >= 0 && ro < colony.grid.length && column >= 0 && column < colony.grid [0].length)
	    {
		if (makenew)
		    colony.grid [ro] [column] = true; // if is adding blocks, set life to true
		else if (!makenew)
		    colony.grid [ro] [column] = false; // if is deleting blocks, set life to false
	    }
	}
    }
}


class Colony
{
    public boolean grid[] [];
    public int generations = 0; // generation count

    public Colony (double density)
    {
	grid = new boolean [125] [200];

	for (int row = 0 ; row < grid.length ; row++)
	    for (int col = 0 ; col < grid [0].length ; col++)
		grid [row] [col] = Math.random () < density; // fill spaces at a certain density
    }


    public Colony (double density, String text) // special glider script
    {
	this (density);

	grid [5] [1] = true;
	grid [5] [2] = true;
	grid [6] [1] = true;
	grid [6] [2] = true;

	grid [5] [11] = true;
	grid [6] [11] = true;
	grid [7] [11] = true;
	grid [4] [12] = true;
	grid [8] [12] = true;
	grid [3] [13] = true;
	grid [3] [14] = true;
	grid [9] [13] = true;
	grid [9] [14] = true;
	grid [6] [15] = true;
	grid [4] [16] = true;
	grid [8] [16] = true;
	grid [5] [17] = true;
	grid [6] [17] = true;
	grid [7] [17] = true;
	grid [6] [18] = true;

	grid [3] [21] = true;
	grid [4] [21] = true;
	grid [5] [21] = true;
	grid [3] [22] = true;
	grid [4] [22] = true;
	grid [5] [22] = true;
	grid [2] [23] = true;
	grid [6] [23] = true;
	grid [2] [25] = true;
	grid [6] [25] = true;
	grid [1] [25] = true;
	grid [7] [25] = true;

	grid [3] [35] = true;
	grid [3] [36] = true;
	grid [4] [35] = true;
	grid [4] [36] = true;

    }


    public void advance ()
    {
	boolean[] [] newarray = new boolean [grid.length] [grid [0].length];

	for (int row = 0 ; row < grid.length ; row++)
	    for (int col = 0 ; col < grid [0].length ; col++)
	    {
		newarray [row] [col] = live (row, col); // make new array and copy each refreshed value over.
	    }
	grid = newarray; // replace old array with new
	generations++; // advance generation count
    }


    public void show (Graphics g)
    {
	for (int row = 0 ; row < grid.length ; row++)
	    for (int col = 0 ; col < grid [0].length ; col++)
	    {
		if (grid [row] [col]) // life
		    g.setColor (Color.black);
		else
		    g.setColor (Color.white);
		g.fillRect (col * 4 + 90, row * 4 + 40, 3, 3);   // draw life form
	    }
    }


    public String importer (String fileName, int choice)  // Can import two types of files, .life files with linebreaks, .'s and *'s AND files with row and column
    {
	String returner = "";
	if (choice == JOptionPane.YES_OPTION) // for examples of files this portion can import (they are prefixed .life), go to http://radicaleye.com/lifepage/patterns/gliderless/gliderless.html
	{
	    for (int row = 0 ; row < grid.length ; row++)
		for (int col = 0 ; col < grid [0].length ; col++)
		    grid [row] [col] = false;
	    try
	    {
		BufferedReader br = new BufferedReader (
			new InputStreamReader (new FileInputStream (fileName)));
		try
		{
		    String line; // line for each line in the file
		    int row = 0;

		    while ((line = br.readLine ()) != null) //lines not empty
		    {
			for (int x = 0 ; x < line.length () ; x++)
			{
			    if (line.charAt (x) == '*') // for every star, it is a true life, so value set to true
				grid [row + 20] [x + 10] = true; //shifts the picture to the right and down so it doesnt start at the edge of grid
			}
			row++;
		    }
		}
		catch (NullPointerException e)
		{
		}
		br.close (); // close bufferedreader
	    }
	    catch (IOException e)
	    {
		returner = "File not Found."; // return error message
	    } 
	}

	else
	{
	    for (int row = 0 ; row < grid.length ; row++)
		for (int col = 0 ; col < grid [0].length ; col++)
		    grid [row] [col] = false;
	    try
	    {
		BufferedReader br = new BufferedReader (
			new InputStreamReader (new FileInputStream (fileName)));
		try
		{
		    String line;  // line for each line in the file
		    String row, column;

		    int comma, int1, int2;
		    while ((line = br.readLine ()) != null) //lines not empty
		    {
			comma = line.indexOf (','); // comma separates row and column
			if (comma != -1)
			{
			    row = line.substring (0, comma);
			    column = line.substring (comma + 1, line.length ());

			    int1 = Integer.parseInt (row); // makes text into number
			    int2 = Integer.parseInt (column);
			    grid [int1] [int2] = true; // make row.column true
			}
		    }
		}
		catch (NullPointerException e)
		{
		}
		br.close (); // close bufferedreader
	    }
	    catch (IOException e)
	    {
		returner = "File not found."; // error message
	    }
	}
	return returner; // return error message
    }


    public void exporter (String fileName, String designName)
    {
	try
	{
	    BufferedWriter out = new BufferedWriter (new FileWriter (fileName));

	    if (designName.equals ("")) // first line of file is name of design, may be custom
		designName = "Untitled";

	    out.write ("" + designName + "\n"); //prints design name
	    out.newLine ();

	    for (int row = 0 ; row < grid.length ; row++)
	    {
		for (int col = 0 ; col < grid [0].length ; col++)
		{
		    if (grid [row] [col] == true) // goes thru array, prints every row/column combination that is true
		    {
			out.write ("" + row + "," + col + "\n");
			out.newLine (); 
		    }
		}
	    }
	    out.close (); // close bufferedwriter
	}

	catch (IOException e)
	{
	}
    }


    public boolean live (int row, int column)
    {
	int liveneighbours = 0;

	if (column - 1 >= 0)// column to the left must exist
	{
	    if (grid [row] [column - 1] == true)
		liveneighbours++;
	    if (row - 1 >= 0)
	    {
		if (grid [row - 1] [column - 1] == true)
		    liveneighbours++;
	    }
	    if (row + 1 < grid.length)
	    {
		if (grid [row + 1] [column - 1] == true)
		    liveneighbours++;
	    }
	}
	if (row - 1 >= 0) // row above must exist
	{
	    if (grid [row - 1] [column] == true)
		liveneighbours++;
	}
	if (row + 1 < grid.length) // row below must exist
	{
	    if (grid [row + 1] [column] == true)
		liveneighbours++;
	}
	if (column + 1 < grid [0].length) // column to the right must exist
	{
	    if (grid [row] [column + 1] == true)
		liveneighbours++;
	    if (row - 1 >= 0)
	    {
		if (grid [row - 1] [column + 1] == true)
		    liveneighbours++;
	    }
	    if (row + 1 < grid.length)
	    {
		if (grid [row + 1] [column + 1] == true)
		    liveneighbours++;
	    }
	}

	if (grid [row] [column] == true)  //life exists
	{
	    if (liveneighbours >= 2 && liveneighbours <= 3) // RULES
		return true;
	    else
		return false;
	}

	else // life = false
	{
	    if (liveneighbours == 3) //RULES
		return true;
	    else
		return false;

	}
    }
}


